#include "denkou.h"
#ifndef __CYGWIN__
#include <sys/mman.h>
#endif

GuFont *guFontNew(char *dirName, char *filename)
{
  GuFont *font = ntMalloc(sizeof(GuFont));
  memset(font, 0, sizeof(GuFont));

  if(dirName == NULL)
    font->filename = ntStringNew(filename);
  else {
    char *p;
    font->filename = p = ntMalloc(strlen(dirName) + strlen(filename) + 2);
    strcpy(p, dirName); p += strlen(p);
    *p++ = '/';
    strcpy(p, filename);
  }
  return font;
}

void guFontOpen(GuFont *font)
{
  int i,h;
  struct stat stat;
  
  if(font->bOpened) nabort("already opened.");
  h = open(font->filename,O_RDONLY);
  if(h < 0) nabort("%s not found.",font->filename);

  if(fstat(h, &stat) < 0) nabort("fstat failed.");
  font->filesize = stat.st_size;
  if(stat.st_size < 4)    nabort("too small file.");
#ifdef __CYGWIN__
  {
    DWORD n,left;
    char *p;
    HANDLE hFile;

    left = stat.st_size;
    font->map = malloc(left);
    if(font->map == NULL) nabort("memory allocation failed (%ld bytes)",left);
    p = font->map;
    hFile = CreateFile(font->filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 
		       FILE_ATTRIBUTE_NORMAL, NULL);
    if(hFile == INVALID_HANDLE_VALUE)
      nabort("%s cannot be read",font->filename);
    while(left > 0){
      ReadFile(hFile, p, left, &n, NULL);
      left -= n;
      p += n;
    }
    CloseHandle(hFile);
  }
#else
  font->map = mmap(0, stat.st_size, PROT_READ, MAP_PRIVATE, h, 0);
  if(font->map == MAP_FAILED) nabort("map failed.");
#endif
  
  if(strncmp(font->map,"\001fcp",4)==0){
    font->pcf_toc = (Pcf_Toc *)font->map;
    for(i=0;i<font->pcf_toc->nTables;i++){
      int32 type = font->pcf_toc->tables[i].type;
      int32 fmt = font->pcf_toc->tables[i].format;
      uchar *sadr = font->map + font->pcf_toc->tables[i].offset;
      
      if(type == PCF_BDF_ENCODINGS)
	font->encode = (Pcf_Encodings *)sadr;
      else if(type == PCF_METRICS){
	if(Is_CMetrics(fmt))
	  font->cmetric = (Pcf_CMetric_Sec *)sadr;
	else
	  font->metric  = (Pcf_Metric_Sec *)sadr;
      }
      else if(type == PCF_BITMAPS){
	int glyph,scan,f;
	Pcf_Bitmaps_Sec *bitmaps;
	font->bitmaps = bitmaps = (Pcf_Bitmaps_Sec *)sadr;
	font->bitmap_base=
	  (byte8 *) &bitmaps->bitmapOffsets[bitmaps->nBitmaps+4];
	f = font->bitmaps->format;
	scan  = (f >> 4) & 3;
	glyph = f & 3;
	font->bytePerData = 1 << scan;
	font->lineAlign   = 1 << glyph;
	font->bMsbFirst   = f & 0x08 ? 1 : 0;
	font->bBigEndian  = f & 0x04 ? 1 : 0;
      }
    }
  }
  else nabort("unknown file format: only pcf supported.");

  if(font->encode == NULL)
    nabort("Encodings section not found in %s",font->filename);
  if(font->metric == NULL && font->cmetric == NULL)
    nabort("Metric section not found in %s",font->filename);
  if(font->bitmaps == NULL)
    nabort("Bitmaps section not found in %s",font->filename);
  font->bOpened = 1;
}

void guFontClose(GuFont *font)
{
  if(!font->bOpened) nabort("not opened.");
#ifdef __CYGWIN__
  free(font->map);
#else
  munmap(font->map, font->filesize);
#endif
  close(font->fontFd);
  font->bOpened = 0;
}

void guFontDump(GuFont *font,unsigned long code)
{
  int i,j,n,bpl,bpc;
  FontMetric mt;
  byte8 *bitmap; 
  if(!font->bOpened) guFontOpen(font);
  n = guFontEncode(font,code);
  printf("code:0x%04lx -> %d\n",code, n);
  guFontMetric(font, n, &mt);

  printf("lbear     : %3d\n", mt.lbear);
  printf("rbear     : %3d\n", mt.rbear);
  printf("width     : %3d\n", mt.width);
  printf("ascent    : %3d\n", mt.ascent);
  printf("descent   : %3d\n", mt.descent);
  printf("attributes: %04x\n",mt.attributes);

  bpl = (mt.rbear - mt.lbear + 7)/8;
  bpl = (bpl + font->lineAlign - 1)/font->lineAlign * font->lineAlign;
  bpc = bpl * (mt.ascent + mt.descent);

  printf("bpl: %3d bytes\n", bpl);
  printf("bpc: %3d bytes\n", bpc);
  guFontBitmap(font, n, &bitmap);
  for(i=0;i<bpc;i+=bpl){
    for(j=0;j<bpl;j++){
      int m,d;
      d = bitmap[i+j];
      for(m=1;m<0x100;m<<=1){
	putchar(d & m ? '*':'.');
      }
    }
    putchar('\n');
  }
}

int guFontEncode(GuFont *font, unsigned long code)
{
  int n,w,col,row;
  Pcf_Encodings *enc = font->encode;
  
  assert(enc != NULL);
  
  col = code & 0xff;
  row = code >> 8;
  if(col < enc->firstCol || col > enc->lastCol ||
     row < enc->firstRow || row > enc->lastRow)
    return enc->defaultCh;
  else {
    w = enc->lastCol - enc->firstCol + 1;
    n = (row - enc->firstRow) * w + col - enc->firstCol;
    return enc->encodings[n];
  }
}

void guFontMetric(GuFont *font,int n, FontMetric *pFm)
{
  if(font->metric!=NULL){
    if(n < 0 || n >= font->metric->nMetrics)
      nabort("index:%d out of range (max:%ld)",
		      n,font->metric->nMetrics);
    memcpy(pFm, &(font->metric->metrics[n]), sizeof(*pFm));
  }
  else if(font->cmetric!=NULL){
    struct _pcf_cmetric_table *pSfm = &(font->cmetric->cmetrics[n]);
    if(n < 0 || n >= font->cmetric->nMetrics)
      nabort("index:%d out of range (max:%d)",
		      n,font->cmetric->nMetrics);
    pFm->lbear      = (int16)pSfm->lbear   - 0x80;
    pFm->rbear      = (int16)pSfm->rbear   - 0x80;
    pFm->width      = (int16)pSfm->width   - 0x80;
    pFm->ascent     = (int16)pSfm->ascent  - 0x80;
    pFm->descent    = (int16)pSfm->descent - 0x80;
    pFm->attributes = 0;
  }
  else nabort("font has no metrics");
}

void guFontBitmap(GuFont *font,int idx,byte8 **pData)
{
  byte8 *adr;
  
  if(idx < 0 || idx >= font->bitmaps->nBitmaps)
    nabort("index:%d out of range. max:%ld",idx,font->bitmaps->nBitmaps);
  adr = font->bitmap_base + font->bitmaps->bitmapOffsets[idx];
  *pData = adr;
}

void guFontGlyph(GuFont *font,unsigned long code,GuGlyph *glyph)
{
  int a,bpl,n;
  FontMetric mt;
  
  if(!font->bOpened) guFontOpen(font);
  n = guFontEncode(font,code);
  guFontMetric(font, n, &mt);

  glyph->lbear   = mt.lbear;
  glyph->rbear   = mt.rbear;
  glyph->width   = mt.width;
  glyph->ascent  = mt.ascent;
  glyph->descent = mt.descent;
  a = font->lineAlign;
  bpl = (mt.rbear - mt.lbear + 7)/8;
  glyph->bpl = (bpl + a - 1)/ a * a;
  
  guFontBitmap(font, n, &glyph->bitmap);
}

void guGlyphDump(GuGlyph *glyph)
{
  int h = glyph->ascent + glyph->descent;
  int x,y;
  byte8 *p;

  p = glyph->bitmap;
  for(y=0;y<h;y++){
    for(x=0;x<glyph->bpl;x++)
      fprintf(stderr,"%02x",*p++);
    fprintf(stderr,"\n");
  }
  fprintf(stderr,"\n");
}
