#include "denkou.h"
#include <ctype.h>

static int get_integer(const char *str, const char *name,
		       const char *rcfile, int line)
{
  int d;

  if(sscanf(str,"%d",&d) < 1)
    nabort("%s:%d Bad %s value:\"%s\" found, it shoud be integer",
	   rcfile, line, name, str);
  return d;
}

static int get_boolean(const char *str, const char *name,
		       const char *rcfile, int line)
{
  int b = 0;

  if(strcmp(str,"1")==0 ||
     strcasecmp(str,"true")==0 ||
     strcasecmp(str,"yes")==0 ||
     strcasecmp(str,"on")==0) b = 1;
  else if(strcmp(str,"0")==0 ||
	  strcasecmp(str,"false")==0 ||
	  strcasecmp(str,"no")==0 ||
	  strcasecmp(str,"off")==0) b = 0;
  else
    nabort("%s:%d Bad %s value:\"%s\" found, it shoud be 0|1|on|off|true|false",
	   rcfile, line, name, str);
  return b;
}

int read_rcfile(dcont_t *dcont, const char *dir, const char *rcfile)
{
  FILE *ifp;
  int len,lc;
  char *p,*p0,*rc_path;
  char line[256],var[256],value[256];

  if(dir == NULL || rcfile[0] == '/')
    rc_path = ntStringNew(rcfile);
  else {
    len = strlen(dir) + strlen(rcfile) + 2;
    rc_path = ntMalloc(len);
    p = rc_path;
    strcpy(p, dir); p+=strlen(p);
    *p++ = '/';
    strcpy(p, rcfile);
  }
  
  ifp = fopen(rc_path,"r");
  if(ifp != NULL){
    if(f_v) fprintf(stderr,"read %s\n",rc_path);
    
    lc = 0;
    while(fgets(line, sizeof(line),ifp)!=NULL){
      lc++;
      for(p=line;isspace(*p);p++);
      if(*p == 0 || *p == '#') continue;
      for(p0 = p;isalnum(*p) || *p=='_';p++);
      len = p - p0;
      strncpy(var,p0,len);
      var[len] = 0;
      while(isspace(*p)) p++;
      if(*p != '=') nabort("Syntax error at %s:%d %s\n",rc_path, lc, p);
      for(p++;isspace(*p);p++);
      for(p0 = p;*p && !isspace(*p);p++);
      len = p - p0;
      strncpy(value,p0,len);
      value[len] = 0;

      if(strcmp(var,"font_dir")==0)
	dcont->font_dir = ntStringNew(value);
      else if(strcmp(var,"fonta")==0)
	dcont->font1_name = ntStringNew(value);
      else if(strcmp(var,"fontk")==0)
	dcont->font2_name = ntStringNew(value);
      else if(strcmp(var,"encoding")==0){
	if(strcasecmp(value,"euc")==0)
	  dcont->encoding = Enc_EUC;
	else if(strcasecmp(value,"sjis")==0||
		strcasecmp(value,"shift_jis")==0)
	  dcont->encoding = Enc_SJIS;
	else nabort("%s:%d Unknown encoding value:\"%s\" found. it should be euc|sjis",
		    rc_path,lc,value);
      }
      else if(strcmp(var,"speed")==0)
	dcont->div = get_integer(value, "speed", rc_path, lc);
      else if(strcmp(var,"no_scroll")==0)
	dcont->no_scroll = get_boolean(value, "no_scroll", rc_path, lc);
      else if(strcmp(var,"loop")==0)
	dcont->b_loop = get_boolean(value, "loop", rc_path, lc);
      else if(strcmp(var,"vertical")==0)
	dcont->b_vertical = get_boolean(value, "vertical", rc_path, lc);
      else if(strcmp(var,"vertical_ank")==0)
	dcont->b_vertical_ank = get_boolean(value, "vertical_ank", rc_path, lc);
      else if(strcmp(var,"led_reverse")==0)
	dcont->b_led_reverse = get_boolean(value, "led_reverse", rc_path, lc);
      else if(strcmp(var,"cmodule")==0)
	dcont->cmodule = get_integer(value, "cmodule", rc_path, lc);
      else if(strcmp(var,"bright")==0)
	dcont->bright = get_integer(value, "bright", rc_path, lc);
      else if(strcmp(var,"tty")==0)
	dcont->scont.tty_name = ntStringNew(value);
      else nabort("%s:%d Unknown parameter \"%s\" found.\n",rc_path,lc,var);
    }
    fclose(ifp);
  }
  else {
    if(f_v) fprintf(stderr,"%s not exist\n",rc_path);
  }

  ntFree(rc_path);
  
  return ifp ? 1 : 0;
}

void read_rcfiles(dcont_t *dcont)
{
  char *home;
  char wd[512];
  
  if((home = getenv("HOME")) != NULL)
    read_rcfile(dcont,home, RCFILE);

  if(getcwd(wd, sizeof(wd)))
    read_rcfile(dcont, wd, RCFILE);
  else
    read_rcfile(dcont, ".", RCFILE);
}
